/**
 * Based on:
 *
 * - https://github.com/threepointone/glamor/blob/78ffd84bfba187f4037a29b389e5a349dbb6e9a8/packages/glamor/src/StyleSheet.ts
 * - https://github.com/styled-components/styled-components/blob/94df8bb57c8ba01cd33268b65df93499a7961cb2/packages/styled-components/src/sheet/Tag.js
 */
/* eslint-disable no-plusplus */
import * as React from 'react';
import { warnOnce } from './logger';
import { NullRuleManager, OptimizedRuleManager, } from './RuleManager';
const MAX_RULE_COUNT = 0xffff;
function createAndMountStyleElement() {
    const el = document.createElement('style');
    el.dataset.glaze = '';
    return document.head.appendChild(el);
}
function getSheet() {
    // Hydrate existing style node if available
    for (let i = 0, len = document.styleSheets.length; i < len; ++i) {
        const styleSheet = document.styleSheets[i];
        if (styleSheet.ownerNode.dataset.glaze === '') {
            return styleSheet;
        }
    }
    const styleEl = createAndMountStyleElement();
    // Avoid Edge bug where empty style elements don't create sheets
    styleEl.appendChild(document.createTextNode(''));
    // Avoid Firefox quirk where the style element might not have a sheet property
    return styleEl.sheet || getSheet();
}
function getInitialRuleIndexes(rules) {
    const ruleIndexesByClassName = new Map();
    for (let i = 0, { length } = rules; i < length; ++i) {
        const rule = rules[i];
        if (rule.type === CSSRule.STYLE_RULE) {
            // Remove leading '.' from class selector
            const className = rule.selectorText.slice(1);
            ruleIndexesByClassName.set(className, i);
        }
        // TODO: Add support for `CSSMediaRule`
    }
    return ruleIndexesByClassName;
}
export class NullStyleInjector {
    constructor() {
        this.ruleManager = new NullRuleManager();
    }
    // eslint-disable-next-line class-methods-use-this
    addRule() {
        return 0;
    }
    // eslint-disable-next-line class-methods-use-this
    nullifyRule() { }
}
export class VirtualStyleInjector {
    constructor() {
        this.ruleManager = new OptimizedRuleManager(this, new Map());
        this.cssTexts = [];
    }
    getStyleElement() {
        return (React.createElement("style", { "data-glaze": "", 
            // eslint-disable-next-line react/no-danger
            dangerouslySetInnerHTML: { __html: this.cssTexts.join('') } }));
    }
    addRule(cssText) {
        return this.cssTexts.push(cssText) - 1;
    }
    // No styles are revoked during SSR
    // eslint-disable-next-line class-methods-use-this
    nullifyRule() { }
}
export class OptimizedStyleInjector {
    constructor() {
        this.sheet = getSheet();
        this.ruleCount = 0;
        this.freeIndexes = [];
        this.ruleManager = new OptimizedRuleManager(this, getInitialRuleIndexes(this.sheet.rules));
    }
    addRule(cssText) {
        if (this.freeIndexes.length) {
            // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
            return this.replaceRule(this.freeIndexes.pop(), cssText);
        }
        return this.sheet.insertRule(cssText, this.ruleCount++);
    }
    nullifyRule(index) {
        if (index === this.ruleCount - 1) {
            this.sheet.deleteRule(--this.ruleCount);
        }
        else {
            // Only allow replacements to prevent modification of existing indexes
            const dummyRule = '#_{}';
            this.freeIndexes.push(this.replaceRule(index, dummyRule));
        }
    }
    replaceRule(index, cssText) {
        this.sheet.deleteRule(index);
        return this.sheet.insertRule(cssText, index);
    }
}
export class DebuggableStyleInjector {
    constructor() {
        this.ruleManager = new OptimizedRuleManager(this, new Map());
        this.nodes = [];
        this.freeIndexes = [];
        this.styleEl = createAndMountStyleElement();
    }
    addRule(cssText) {
        if (this.freeIndexes.length) {
            // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
            return this.replaceRule(this.freeIndexes.pop(), cssText);
        }
        const index = this.nodes.length;
        if (index === MAX_RULE_COUNT) {
            warnOnce('An excessive amount of styling rules have been injected at runtime. Consider using inline styles or CSS variables instead.');
        }
        const node = document.createTextNode(cssText);
        this.styleEl.appendChild(node);
        this.nodes[index] = node;
        return index;
    }
    nullifyRule(index) {
        if (index === this.nodes.length - 1) {
            // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
            this.styleEl.removeChild(this.nodes.pop());
        }
        else {
            // Only allow replacements to prevent modification of existing indexes
            const dummyRule = '';
            this.freeIndexes.push(this.replaceRule(index, dummyRule));
        }
    }
    replaceRule(index, cssText) {
        this.nodes[index].textContent = cssText;
        return index;
    }
}
//# sourceMappingURL=StyleInjector.js.map